# This script joins called LRs

# import argparse
from multiprocessing import Value,Pool
from typing import Type
from xml.dom.minidom import Attr
import pandas
import glob
import sys
import statistics
import traceback
from tqdm import tqdm
import logging

class JoinLR():

    def __init__(self,ds,
                 args):
        self.th=args.threads
        self.log_path = args.workDir+'/elarodon.log'
        logging.basicConfig(level=logging.DEBUG, filename=self.log_path, filemode="a",
                            format="%(asctime)s %(funcName)s %(lineno)d %(message)s")
        p=Pool(self.th)
        self.maxDistToJoin=args.maxDistToJoin
        self.allFusions={}
        self.allInsertions={}
        for res in tqdm(p.imap_unordered(self.processFile,ds,
                                         chunksize=10),
                        total=len(ds)):
            fileFusions,fileInsertions=res
            if len(fileFusions)>0:
                for chrom,fusionToInfo in fileFusions.items():
                    if chrom not in self.allFusions.keys():
                        self.allFusions[chrom]=fusionToInfo
                    else:
                        for fusion,info in fusionToInfo.items():
                            if fusion not in self.allFusions[chrom].keys():
                                self.allFusions[chrom][fusion]=info
                            else:
                                for i in [0,2,4,5,6,7,8,9,10]:
                                    self.allFusions[chrom][fusion][i].extend(info[i])
                                    # Error percents
                                for jointType,jointSeqToNum in enumerate(info[1]):
                                    for jointSeq,num in jointSeqToNum.items():
                                        if jointSeq not in self.allFusions[chrom][fusion][1][jointType].keys():
                                            self.allFusions[chrom][fusion][1][jointType][jointSeq]=num
                                        else:
                                            self.allFusions[chrom][fusion][1][jointType][jointSeq]+=num
            if len(fileInsertions)>0:
                for chrom1_start,info in fileInsertions.items():
                    if chrom1_start not in self.allInsertions.keys():
                        self.allInsertions[chrom1_start]=info
                    else:
                        for insInfo in info:
                            self.allInsertions[chrom1_start].append(insInfo)

    def processFile(self,d):
        allFusions={}
        allInsertions={}
        if 'fusions.csv' in d:
            try:
                data=pandas.read_csv(d,sep='\t',
                                     header=0,
                                     keep_default_na=False,
                                     low_memory=False,
                                     encoding='utf-8')
            except UnicodeDecodeError:
                print('ERROR (6)! Could not read the following file (UnicodeDecodeError):')
                print(d)
                exit(6)
            for i,row in data.iterrows():
                row = row.to_list()
                fusion=row[:8]
                # Compare chromosomes
                fusionOrdered=fusion[:]
                try:
                    chrom1=int(fusion[0].replace('chr',''))
                except TypeError:
                    chrom1=fusion[0]
                except ValueError:
                    chrom1=fusion[0]
                try:
                    chrom2=int(fusion[4].replace('chr',''))
                except TypeError:
                    chrom2=fusion[4]
                except ValueError:
                    chrom2=fusion[4]
                except AttributeError:
                    print('AttribureError:',fusion[4])
                    print(d)
                    print(row)
                    print(fusion)
                    exit()
                readLen1=row[10]
                readLen2=row[11]
                # Change the order only if it is not T, D, or I
                if (row[3] not in ['T','D','I','V'] and
                    'D' not in row[3]):
                    if ((type(chrom1)==int and
                         type(chrom2)==int) or
                        (type(chrom1)==str and
                          type(chrom2)==str)):
                        # If chrom2 is less than chrom1
                        if chrom2<chrom1:
                            # We change their order
                            fusionOrdered=[*fusion[4:8],
                                           *fusion[0:4]]
                            readLen1=row[11]
                            readLen2=row[10]
                        # If pos2<pos1, we also change the order
                        elif (chrom1==chrom2 and 
                              checkPosOrder(fusion[1],
                                            fusion[5],
                                            fusion[3])):
                            fusionOrdered=[*fusion[4:8],
                                           *fusion[0:4]]
                            readLen1=row[11]
                            readLen2=row[10]
                    # If the 1st chromosome is str and 2nd is int
                    elif (type(chrom1)==str and
                          type(chrom2)==int):
                        # We also change their order
                        fusionOrdered=[*fusion[4:8],
                                       *fusion[0:4]]
                        readLen1=row[11]
                        readLen2=row[10]
                errPercs=str(row[12]).split(',')
                # Start_End_Locations,Number_of_parts
                startEndLoc=row[16]
                partsNum=int(row[17])
                # SeqBefore,SeqAfter
                seqBefore=row[18]
                seqAfter=row[19]
                fusionOrdered=tuple(fusionOrdered)
                chrom=fusionOrdered[0]
                if type(fusionOrdered[1])==int:
                    pos=fusionOrdered[1]
                else:
                    pos=int(fusionOrdered[1].split(',')[0])
                # Now save it by the 1st chromosome
                if chrom not in allFusions.keys():
                    allFusions[chrom]={}
                if fusionOrdered not in allFusions[chrom].keys():
                    # [reads_names,[jointType1,jointType2,jointType3],
                    # [MQs],refCov,[readLens1],[readLens2],[errPercs]
                    # [First_Last_Part],[Num_of_Parts],[Seq_before],[Seq,after]]
                    allFusions[chrom][fusionOrdered]=[[],[{},{},{},{},{}],
                                                           [],row[9],[],[],[],
                                                           [],[],[],[]]
                if row[20] not in allFusions[chrom][fusionOrdered][0]:
                    # read_name
                    allFusions[chrom][fusionOrdered][0].append(row[20])
                    # MQ
                    allFusions[chrom][fusionOrdered][2].append(row[8])
                    # Read lengths
                    allFusions[chrom][fusionOrdered][4].append(str(readLen1))
                    allFusions[chrom][fusionOrdered][5].append(int(readLen2))
                    # Error percents
                    allFusions[chrom][fusionOrdered][6].extend(errPercs)
                    # Parts of reads (numbers)
                    allFusions[chrom][fusionOrdered][7].append(str(startEndLoc))
                    allFusions[chrom][fusionOrdered][8].append(str(partsNum))
                    # Seq before and after BND
                    allFusions[chrom][fusionOrdered][9].append(seqBefore)
                    allFusions[chrom][fusionOrdered][10].append(seqAfter)
                    # If the sequence of jointType is absent
                    if (row[15]=='' or
                        type(row[15])==float):
                        # Then we save its length
                        jointSeq=str(row[14])
                    else:
                        jointSeq=row[15]
                    # Determine, to which list we should save the joint info
                    for jointType,jointSeq in zip(str(row[13]).split(','),
                                                  str(row[14]).split(',')):
                        jointNum=['inter','scarless','newseq','cigar','inversion'].index(jointType)
                        if jointSeq not in allFusions[chrom][fusionOrdered][1][jointNum].keys():
                            allFusions[chrom][fusionOrdered][1][jointNum][jointSeq]=0
                        allFusions[chrom][fusionOrdered][1][jointNum][jointSeq]+=1
            del(data)
            return(allFusions,allInsertions)
        if 'insertions.csv' in d:
            data=pandas.read_csv(d,sep='\t')
            sorted_data = data.sort_values(by='Pos1')

            miniGroups = []
            miniGroupInds = {}

            miniGroupSeqs = {}
            
            for i,row in sorted_data.iterrows():
                row = row.to_list()
                
                ins_added = False 
                # print(row)

                # row
                # 0 - chrom, 1 - pos, 2 - len ins, 3 - sequence, 4 - read number, 5 - total cov, 6 - MQ, 7 - Read Muts1, 8 - read name 
                
                if miniGroups != []:
                    for ind_last_group, last_ins in sorted(miniGroupInds.items(), key=lambda x: x[0], reverse=True):
                        # print('ind_last_group', ind_last_group)
                        if (ins_added == False and 
                            row[0] == last_ins[0] and
                            abs(int(row[1]) - int(last_ins[1])) <= 100 and
                            abs(int(row[2]) - int(last_ins[2])) <= 100):
                            
                            # minigroup
                            # 0 - chrom, 1 - pos, 2 - len, 3 - total cov, 4 - MQ, 5 - Read_Muts1, 6 - read name

                            miniGroups[ind_last_group][0].append((row[0], # chrom 0
                                                                int(row[1]), # pos 1
                                                                int(row[2]), # len 2
                                                                int(row[5]), # tot cov 3
                                                                int(row[6]), # MQ 4
                                                                float(row[7]), # RM1 5
                                                                row[8])) # RN 6
                            miniGroupInds[ind_last_group] = [row[0], int(row[1]), int(row[2])] # add new last ins in group
                            
                            if row[3] not in miniGroups[ind_last_group][1]:
                                miniGroups[ind_last_group][1][row[3]] = 0                    
                            miniGroups[ind_last_group][1][row[3]] += 1
                            # print('We add INS to ', ind_last_group, ' group')
                            # print(miniGroups[ind_last_group])
                            # print(miniGroupInds[ind_last_group])
                            # input()
                            ins_added = True
                    
                else:
                    # minigroups and minigroup sequences
                    miniGroups.append([[], {}])
                    miniGroups[-1][0].append((row[0], # chrom 0
                                            int(row[1]), # pos 1
                                            int(row[2]), # len 2
                                            int(row[5]), # tot cov 3
                                            int(row[6]), # MQ 4
                                            float(row[7]), # RM1 5
                                            row[8])) # RN 6
                    miniGroupInds[len(miniGroups)-1] = [row[0], int(row[1]), int(row[2])]

                    if row[3] not in miniGroups[-1][1]:
                        miniGroups[-1][1][row[3]] = 0                    
                    miniGroups[-1][1][row[3]] += 1
                    ins_added = True
                
                if ins_added == False:

                    miniGroups.append([[], {}])
                    miniGroups[-1][0].append((row[0], # chrom 0
                                            int(row[1]), # pos 1
                                            int(row[2]), # len 2
                                            int(row[5]), # tot cov 3
                                            int(row[6]), # MQ 4
                                            float(row[7]), # RM1 5
                                            row[8])) # RN 6
                    miniGroupInds[len(miniGroups)-1] = [row[0], int(row[1]), int(row[2])]

                    if row[3] not in miniGroups[-1][1]:
                        miniGroups[-1][1][row[3]] = 0                    
                    miniGroups[-1][1][row[3]] += 1
                
                # print('ins_added', ins_added)
                # print('miniGroups', miniGroups)
                # input()

            for miniGroup in miniGroups:

                # print('miniGroup', miniGroup)
                # input()

                chrom1_start = miniGroup[0][0][0]
                pos1_start = miniGroup[0][0][1]
                total_cov_middle = round(sum([x[3] for x in miniGroup[0]]) / len(miniGroup[0]))      
                len_ins_middle = round(sum([x[2] for x in miniGroup[0]]) / len(miniGroup[0]))
                read_muts = [str(x[5]) for x in miniGroup[0]]
                read_names = [x[6] for x in miniGroup[0]]
                mqs = [x[4] for x in miniGroup[0]]

                miniGroupSeqs = miniGroup[1]
                
                if chrom1_start not in allInsertions.keys():
                    allInsertions[chrom1_start]=[]
                allInsertions[chrom1_start].append([pos1_start,
                                                    read_names, 
                                                    miniGroupSeqs,
                                                    total_cov_middle,
                                                    mqs,
                                                    read_muts])
        del(miniGroups)
        del(data)
        return(allFusions,allInsertions)

    def joinAllSimilarLRs(self):
        print('Joining similar fusions...')
        p=Pool(self.th)
        self.allFusionsJoined1={}
        for res in tqdm(p.imap_unordered(self.joinChromSimilarLRs,
                                         self.allFusions.keys(),
                                         chunksize=10),
                        total=len(self.allFusions.keys())):
            self.allFusionsJoined1[res[0]]=res[1]
        

    def joinChromSimilarLRs(self,chrom):
        chromFusionsJoined1={}
        # Sort by 1st coordinate
        fusionsSorted=sorted(self.allFusions[chrom],
                             key=lambda item:[item[0],
                                              int(str(item[1]).split(',')[0]), # pos
                                              item[4],
                                              int(str(item[5]).split(',')[0])])
        # Stores fusion IDs that have been already joined to another fusion
        alreadyJoinedFusionIDs=set()
        # Join fusions by searching for the first coordinate
        for i,fusion in enumerate(fusionsSorted):
            if (i==len(fusionsSorted)-1 or
                i in alreadyJoinedFusionIDs):
                continue
            joinedNum=self.allFusions[chrom][fusion]
            joinedFusion=fusion[:]
            fusionNum=i+1
            # Compare coordinates of the current fusion and several further
            while(True):
                if fusionNum==len(fusionsSorted)-1:
                    break
                nextFusion=fusionsSorted[fusionNum]
                # If chromosomes are the same
                if (fusion[0]==nextFusion[0] and
                    fusion[4]==nextFusion[4] and
                    # And strands are corresponding
                    # [+,+] vs [+,+] or [-,-] | [-,-] vs [+,+] or [-,-]
                    # [+,-] vs [+,-] or [-,+]
                    (((fusion[2]==fusion[6] and
                       nextFusion[2]==nextFusion[6])) or
                     (fusion[2]!=fusion[6] and
                      nextFusion[2]!=nextFusion[6])) and
                    (fusion[3]==nextFusion[3] and
                     fusion[7]==nextFusion[7]) and
                    # And the coordinates are similar
                    compareCoords(fusion[1],fusion[5],
                                  nextFusion[1],nextFusion[5],
                                  self.maxDistToJoin)):
                    # Join total numbers for the joint fusions
                    joinedNum[0].extend(self.allFusions[chrom][nextFusion][0])
                    joinedNum[2].extend(self.allFusions[chrom][nextFusion][2])
                    joinedNum[4].extend(self.allFusions[chrom][nextFusion][4])
                    joinedNum[5].extend(self.allFusions[chrom][nextFusion][5])
                    joinedNum[6].extend(self.allFusions[chrom][nextFusion][6])
                    joinedNum[7].extend(self.allFusions[chrom][nextFusion][7])
                    joinedNum[8].extend(self.allFusions[chrom][nextFusion][8])
                    joinedNum[9].extend(self.allFusions[chrom][nextFusion][9])
                    joinedNum[9].extend(self.allFusions[chrom][nextFusion][10])
                    # Join numbers for different joint types
                    for typeNum in range(5):
                        # Go through all sequences of the nextFusion
                        for nextFusionSeq,seqNum in self.allFusions[chrom][nextFusion][1][typeNum].items():
                            if nextFusionSeq not in joinedNum[1][typeNum].keys():
                                joinedNum[1][typeNum][nextFusionSeq]=0
                            joinedNum[1][typeNum][nextFusionSeq]+=seqNum
                    alreadyJoinedFusionIDs.add(fusionNum)
                elif (fusion[0]!=nextFusion[0] or
                      abs(int(str(fusion[1]).split(',')[0])-int(str(nextFusion[1]).split(',')[0]))>self.maxDistToJoin):
                    break
                fusionNum+=1
            if joinedFusion not in chromFusionsJoined1.keys():
                chromFusionsJoined1[joinedFusion]=joinedNum
            else:
                chromFusionsJoined1[joinedFusion][0].extend(joinedNum[0])
                for typeNum in range(5):
                    # Go through all sequences of the nextFusion
                    for nextFusionSeq,seqNum in joinedNum[1][typeNum].items():
                        if nextFusionSeq not in chromFusionsJoined1[joinedFusion][1][typeNum].keys():
                            chromFusionsJoined1[joinedFusion][1][typeNum][nextFusionSeq]=0
                        chromFusionsJoined1[joinedFusion][1][typeNum][nextFusionSeq]+=seqNum
                chromFusionsJoined1[joinedFusion][2].extend(joinedNum[2])
        return(chrom,chromFusionsJoined1)

    def writeFusionToOutput(self,outFile):
        print('Writing fusions to output...')
        rFile=open(outFile,'w')
        rFile.write('\t'.join(['Chrom1','Pos1','Strand1','Junction_Side1',
                               'Chrom2','Pos2','Strand2','Junction_Side2',
                               'Read_Number','Ref_Coverage','MQ_median',
                               'Read_Len1','Read_Len2','Read_Muts',
                               'Inter_Joints', 
                               'Scarless_Joints',
                               'NewSeq_Joints',
                               'Cigar_Num','Inversion_Num',
                               'Start_End_Locations','Number_of_parts',
                               'Seqs_before_BND','Seqs_after_BND',
                               'Read_Names'])+'\n')
        sumLen=sum(len(item.keys()) for item in self.allFusionsJoined1.values())
        showPercWork(0,sumLen)
        fusioNum=0
        # Stores GC-content values for LRs of different groups:
        # inter_lambda, equal_lambda, distance_lambda, inter_dif_chroms, equal_dif_chroms, dist_dif_chroms, 
        # inter_same_chrom, equal_same_chrom, dist_same_chrom
        gcContents=[[[],[],[]],
                    [[],[],[]],
                    [[],[],[]]]
        # Stores GC-content values for LRs with different coverage values:
        # [cov==1,cov==2,cov==3...cov==9,cov>=10]
        # It doesn't include LRs with distance
        gcContentForCovs=[[],[],[],
                          [],[],[],
                          [],[],[],[]]
        for chrom in tqdm(sorted(self.allFusionsJoined1.keys()),
                          total=len(self.allFusionsJoined1.keys())):
            for fusion,numbers in self.allFusionsJoined1[chrom].items():
                # Stores length of the joint or its sequence
                jointTypeInfo=[]
                for typeNum in range(5):
                    if typeNum>2:
                        jointTypeInfo.append(str(sum(numbers[1][typeNum].values())))
                        continue
                    jointSeqLens=[]
                    for jointSeqLen,num in numbers[1][typeNum].items():
                        jointSeqLens.extend([str(jointSeqLen)]*num)
                    if len(jointSeqLens)==0:
                        jointTypeInfo.append('0')
                    else:
                        jointTypeInfo.append(','.join(jointSeqLens))
                try:
                    out='\t'.join([fusion[0],str(fusion[1]),
                                   fusion[2],fusion[3],
                                   fusion[4],str(fusion[5]),
                                   fusion[6],fusion[7],
                                   str(len(numbers[0])),
                                   str(int(numbers[3])),
                                   str(statistics.median(numbers[2])),
                                   ','.join(numbers[4]),
                                   str(max(numbers[5])),
                                   ','.join(numbers[6]),
                                   *jointTypeInfo,
                                   ','.join(numbers[7]),
                                   ','.join(numbers[8]),
                                   ','.join(numbers[9]),
                                   ','.join(numbers[10]),
                                   ','.join(numbers[0])])
                except TypeError:
                    print('ERROR (2)! Incorrect type of data for output:')
                    print(jointTypeInfo)
                    print(fusion)
                    print(numbers)
                    exit(2)
                rFile.write(out+'\n')
        rFile.close()
        print()

    def writeInsertionsToOutput(self,outFile):
        # Insertions
        print('Writing insertions to output...')
        rFile=open(outFile,'w')
        rFile.write('\t'.join(['Chrom','Pos','Insertion_Length','Sequence',
                               'Read_Number','Total_coverage','MQ_median',
                               'Read_Names','Read_Muts'])+'\n')
        for chrom,posToInsertions in tqdm(sorted(self.allInsertions.items()),
                                          total=len(self.allInsertions.keys())):
            for insInfo in sorted(posToInsertions, key=lambda x: x[0]):
                pos = insInfo[0]
                insInfo = insInfo[1:]
                outLineList=[str(chrom),
                             str(pos)]
                # Get the most frequent seq
                # [altCov,{insSeqs:1},totalCov,[MQs]]
                for seq,num in sorted(insInfo[1].items(),
                                      key=lambda item:item[1],
                                      reverse=True):
                    outLineList.append(str(len(seq)))
                    outLineList.append(seq)
                    break
                outLineList.extend([str(len(insInfo[0])),
                                    str(insInfo[2]),
                                    str(statistics.median(insInfo[3])),
                                    ','.join(insInfo[0]),
                                    ','.join(insInfo[4])])
                rFile.write('\t'.join(outLineList)+'\n')
        rFile.close()

def showPercWork(done,allWork):
    percDoneWork=round((done/allWork)*100,2)
    sys.stdout.write("\r"+str(percDoneWork)+"%")
    sys.stdout.flush()

def getGC(seq):
    cNum=seq.count('C')
    gNum=seq.count('G')
    gcContent=round((cNum+gNum)*100/len(seq),0)
    return(gcContent)

def compareCoords(pos1,pos2,
                  nextPos1,nextPos2,
                  maxDist):
    pos1=str(pos1).split(',')
    pos2=str(pos2).split(',')
    nextPos1=str(nextPos1).split(',')
    nextPos2=str(nextPos2).split(',')
    if (len(pos1)!=len(nextPos1) or
        len(pos2)!=len(nextPos2)):
        return(False)
    for p1,np1 in zip(pos1,nextPos1):
        if abs(int(p1)-int(np1))>maxDist:
            return(False)
    for p2,np2 in zip(pos2,nextPos2):
        if abs(int(p2)-int(np2))>maxDist:
            return(False)
    return(True)

# Returns True if pos2<pos1
def checkPosOrder(pos1,pos2,juncType='L'):
    if juncType!='L' and juncType!='R':
        return(False)
    pos1=str(pos1).split(',')
    pos2=str(pos2).split(',')
    if int(pos2[0])<int(pos1[0]):
        return(True)
    return(False)

# if __name__=='__main__':

#     # Section of reading arguments
#     par=argparse.ArgumentParser(description='This script joined called LRs')
#     par.add_argument('--input-files','-in',
#                      dest='inFiles',type=str,
#                      help='regular expression for CSV-files with called LRs',
#                      required=True)
#     par.add_argument('--maximal-distance-join','-join',
#                      dest='maxDistToJoin',type=int,
#                      help='maximal acceptable distance of two neighboring fusions to join them. Default: 30',
#                      required=False,default=30)
#     par.add_argument('--output-insertions-file','-outi',
#                      dest='outInsFile',type=str,
#                      help='CSV output file for insertions',
#                      required=True)
#     par.add_argument('--output-file','-out',
#                      dest='outFile',type=str,
#                      help='CSV file for other large rearrangements',
#                      required=True)
#     par.add_argument('--threads','-th',
#                      dest='threads',type=int,
#                      help='number of threads to process files. Default: 8',
#                      required=False,default=8)
#     args=par.parse_args()

#     ds=glob.glob(args.inFiles)
#     if len(ds)==0:
#         print('ERROR (1)! No files were chosen:')
#         print(args.inFiles)
#         exit(1)
#     print('The number of files chosen:',len(ds))

#     print('Reading input files with',args.threads,'threads...')
#     jl=JoinLR(ds,args)
    
#     print('The total number of fusions:',sum(len(item) for item in jl.allFusions.values()))
#     print('The total number of insertions:',sum(len(item) for item in jl.allInsertions.values()))

#     jl.joinAllSimilarLRs()
#     jl.writeFusionToOutput(args.outFile)
#     jl.writeInsertionsToOutput(args.outInsFile)

##    print('Writing GC-content of different LRs...')
##    # Output-file for GC-content of different LRs
##    # Dist lambda | equal lambda | inter lambda...
##    rFileGC=open(args.outFile[:args.outFile.rfind('.')]+'.GC_content.csv','w')
##    rFileGC.write('\t'.join(['Chromosomes',
##                             'GC-content',
##                             'Joint Type'])+'\n')
##    # Go through lambda, dif_chroms, same_chroms
##    for i,chromType in enumerate(['Lambda','Dif_Chrom','Same_Chrom']):
##        for j,readOverType in enumerate(['Inter','Equal','Dist']):
##            for value in gcContents[i][j]:
##                rFileGC.write('\t'.join([chromType,
##                                         str(value),
##                                         readOverType])+'\n')
##    rFileGC.close()
##    print('Writing GC-content for LRs with different coverage values...')
##    # Output-file for GC-content for LRs with different coverage values
##    # Coverage | GC-content
##    rFileGC=open(args.outFile[:args.outFile.rfind('.')]+'.GC_content_vs_cov.csv','w')
##    rFileGC.write('\t'.join(['Coverage',
##                             'GC-content'])+'\n')
##    # Go through all coverage values
##    for i in range(10):
##        for value in gcContentForCovs[i]:
##            if i==9:
##                rFileGC.write('\t'.join(['>=10',
##                                         str(value)])+'\n')
##            else:
##                rFileGC.write('\t'.join([str(i+1),
##                                         str(value)])+'\n')
##    rFileGC.close()



import pysam
import logging

class BamReader():

    def __init__(self, bamFile, ths, log_path):

        self.bamFile = bamFile
        self.threads = ths
        self.log_path = log_path
        logging.basicConfig(level=logging.DEBUG, filename=self.log_path, filemode="a",
                format="%(asctime)s %(funcName)s %(lineno)d %(message)s")

    def read_bam(self):

        bam = pysam.AlignmentFile(self.bamFile, 'rb', threads=self.threads)
        chromosomes = bam.references
        lengths = bam.lengths
        chrom_lengths = dict(zip(chromosomes, lengths))
        logging.info('BAM file was read: '+str(self.bamFile))
        logging.info('Chromosomes: '+str(chromosomes))
        logging.info('Lengths: '+str(lengths))
        
        return chrom_lengths

        
        




import argparse
import subprocess
from collections import defaultdict
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
import os
import shutil
from statistics import median, mean
import pysam
import sys
import pandas as pd
from multiprocessing import Pool, cpu_count
from tqdm import tqdm
import logging

class InsertionProcessor:
    def __init__(self, log_path, ins_file, lr_file, ref_genome, name, sam_file=None, threads=6, not_remove_trash=False):
        self.log_path = log_path
        logging.basicConfig(level=logging.DEBUG, filename=self.log_path, filemode="a",
                            format="%(asctime)s %(funcName)s %(lineno)d %(message)s")
        self.ins_file = ins_file
        self.lr_file = lr_file
        self.ref_genome = ref_genome
        self.name = name
        self.sam_file = sam_file
        self.insertions = {}
        self.tandem_duplications = []
        self.tandem_duplications_ids = []
        self.threads = threads
        self.not_remove_trash = not_remove_trash

    def read_insertions(self):
        print('Reading INS-file...')
        logging.info('Reading INS-file...')
        df = pd.read_csv(self.ins_file, delimiter='\t')
        for index, row in df.iterrows():
            ins_id = index
            if ',' in row['Read_Names']:
                row['Read_Number'] = str(len(row['Read_Names'].split(',')))
            self.insertions[ins_id] = row

    def write_fasta(self, output_file):
        print('Writing FASTA-file...')
        logging.info('Writing FASTA-file...')
        logging.info('output_file: '+output_file)
        with open(output_file, 'w') as fasta_file:
            for ins_id, data in self.insertions.items():
                # print('data', data)
                # input()
                seq_id = str(ins_id)+'*'+data['Chrom']+'*'+str(data['Pos'])+'*'+str(data['Insertion_Length'])
                sequence = SeqRecord(Seq(data['Sequence']), id=seq_id, description='')
                SeqIO.write(sequence, fasta_file, 'fasta')
        # input('STOP FASTA')

    def map_with_minimap2(self, fasta_file):
        print('Mapping...')
        logging.info('Mapping...')
        if self.sam_file and os.path.exists(self.sam_file):
            print("Using existing SAM file...")
            logging.info('"Using existing SAM file...')
            return
        output_sam = self.name + "_mapping_results.sam"
        logging.info('output_sam: '+output_sam)
        subprocess.run([
            'minimap2', '-a', self.ref_genome, fasta_file],
            stdout=open(output_sam, 'w')
        )
        logging.info(' '.join(['minimap2', '-a', self.ref_genome, fasta_file]))
        self.sam_file = output_sam
        # input('STOP mapping')

    def analyze_mapping_results(self):
        print('Searching tandem duplications...')
        logging.info('Searching tandem duplications...')
        with open(self.sam_file, 'r') as sam_file:
            for line in sam_file:
                if line.startswith('@'):
                    continue
                parts = line.split('\t')
                # print('parts', parts)
                # input()
                read_id = parts[0]
                ins_id, chrom, coordinate, length = read_id.split('*')
                ins_id = int(ins_id)
                coordinate = int(coordinate)
                length = int(length)

                map_chrom = parts[2]
                map_pos = int(parts[3])
                if map_chrom == chrom and coordinate - length - 50 < map_pos < coordinate + length + 50:
                    self.tandem_duplications.append((ins_id, map_pos, length))
                    self.tandem_duplications_ids.append(ins_id)
        # print('self.tandem_duplications', self.tandem_duplications)
        # print()
        # input('STOP SAM')

    def merge_tandem_duplications(self):
        print('Merging tandem duplications...')
        logging.info('Merging tandem duplications...')
        self.merged_insertions = defaultdict(list)
        for ins_id_1, map_pos_1, length_1 in self.tandem_duplications:
            for ins_id_2, map_pos_2, length_2 in self.tandem_duplications:
                if abs(map_pos_1 - map_pos_2) < 50 and abs(length_1 - length_2):
                    self.merged_insertions[ins_id_1].append((ins_id_2, map_pos_2, length_2))
                    self.merged_insertions[ins_id_2].append((ins_id_1, map_pos_1, length_1))
        # print('self.merged_insertions', self.merged_insertions)
        # input('STOP merging')

    def copy_file(self, source_path, destination_path):
        try:
            shutil.copy(source_path, destination_path)
            # print(f"File copied from {source_path} to {destination_path}")
        except Exception as e:
            print("Error occurred while copying file: " + e)
            logging.error("Error occurred while copying file: " + e)
    
    def get_sequence_from_genome(self, chrom_name, temp_start, temp_end, diff, flag):
        # attempts = 0
        seq = None
        while (seq is None):
            if flag == 'L':
                try:
                    seq = str(ff.fetch(region=str(chrom_name)+':'+str(temp_start-diff)+'-'+str(temp_end)))
                except:
                    seq = None
                    if diff > 5:
                        diff -= 5
                        if diff < 5:
                            diff = 5
                            continue
            elif flag == 'R':
                try:
                    seq = str(ff.fetch(region=str(chrom_name)+':'+str(temp_start)+'-'+str(temp_end+diff)))
                except:
                    seq = None
                    if diff > 5:
                        diff -= 5
                        if diff < 5:
                            diff = 5
                            continue
            # if attempts == 4:
            #     diff = 2
            # attempts += 1
            if seq is None:
                ff = pysam.FastaFile(self.ref_genome, filepath_index=self.ref_genome+'.fai')
                if diff <= 5:
                    diff -= 1
                    if diff >= 1:
                        continue
                    else:
                        # print()
                        # print('Error! Unable to extract the sequence from the reference genome!')
                        # print(chrom_name, temp_start, temp_end, diff, flag)
                        return ''
                        # sys.exit()
        # if flag == 'R':
        #     print(chrom_name, temp_start, temp_end, diff, flag)
        #     print(seq)
        return (seq.upper())
    
    def process_extraction(self, data_ins):

        left_s = self.get_sequence_from_genome(data_ins['Chrom'], int(data_ins['map_pos']), int(data_ins['map_pos']), 30, 'L')
        right_s = self.get_sequence_from_genome(data_ins['Chrom'], int(data_ins['Pos']), int(data_ins['Pos']), 30, 'R')

        return (data_ins['ID'], left_s, right_s)

    def extract_sequences_before_after(self):
        print('Extracting sequences before and after insertions...')
        logging.info('Extracting sequences before and after insertions...')
        rows = []
        for ins_id, map_pos, length in self.tandem_duplications:  
            self.insertions[ins_id]['ID'] = ins_id
            self.insertions[ins_id]['map_pos'] = ins_id
            rows.append(self.insertions[ins_id])

        # create pool
        num_processes = self.threads
        pool = Pool(processes=num_processes)

        # give single rows to pool
        if len(rows) < 1000:
            chunk_size_val = 100
        elif len(rows) < 10000:
            chunk_size_val = 1000
        else:
            chunk_size_val = 10000

        results = list(tqdm(pool.imap_unordered(self.process_extraction, rows, chunksize=chunk_size_val), total=len(rows), desc="Processing insertions")) #, colour='blue'))
        # results = pool.map(self.process_row_INS, rows)

        pool.close()
        pool.join()

        for result in results:
            id = result[0]
            seq_bef = result[1]
            seq_aft = result[2]
            self.insertions[id]['SBB'] = seq_bef
            self.insertions[id]['SAB'] = seq_aft


    def update_large_rearrangements_file(self):
        print('Updating LRs data...')
        logging.info('Updating LRs data...')
        uniq = []
        with open(self.lr_file[:-4]+'_mapped_ins.csv', 'a') as lr_file:
            for ins_id, map_pos, length in self.tandem_duplications:
                if ins_id not in self.merged_insertions:
                    ins_info = self.insertions[ins_id]
                    # print('ins_info', ins_info)
                    # input()
                    lr_file.write('\t'.join([str(ins_info['Chrom']), # Chrom1
                                            str(map_pos)+','+str(ins_info['Pos']), # Pos1
                                            '+', # Strand1
                                            'D', # Junction_Side1
                                            str(ins_info['Chrom']), # Chrom2
                                            str(map_pos)+','+str(map_pos+length), # Pos2
                                            '+', # Strand2
                                            'D', # Jinction_Side2
                                            str(ins_info['Read_Number']), # Read_Number
                                            str(int(ins_info['Total_coverage']) - int(ins_info['Read_Number'])), # Ref_Coverage
                                            str(ins_info['MQ_median']), # MQ_median
                                            '99', # Read_Len1
                                            '99', # Read_Len2
                                            str(ins_info['Read_Muts']), # Read_Muts
                                            '0', # Inter_Joints
                                            '1', # Scarless_Joints
                                            '0', # NewSeq_Joints
                                            '1', # Cigar_Num
                                            '0', # Inversion_Num
                                            '0', # Start_End_Locations
                                            '0', # Number_of_parts
                                            ','.join(ins_info['SBB']), # Seqs_before_BND
                                            ','.join(ins_info['SAB']), # Seqs_after_BND
                                            ins_info['Read_Names'] # Read_names
                                            ]) + '\n')
                    uniq.append(ins_id)
                else:
                    ins_ids = [ins_id]
                    ins_ids.extend([i for i in self.merged_insertions[ins_id]])

                    ins_info_main = [self.insertions[ins_id]]
                    # print('ins_info_main', ins_info_main)
                    # input()
                    ins_info_main.extend([self.insertions[id[0]] for id in self.merged_insertions[ins_id]])

                    map_poses = [map_pos]
                    map_poses.extend([pos[1] for pos in self.merged_insertions[ins_id]])

                    map_lens = [length]
                    map_lens.extend([l[2] for l in self.merged_insertions[ins_id]])
                    
                    poses = [x['Pos'] for x in ins_info_main]
                    chroms = [x['Chrom'] for x in ins_info_main]
                    rn = [int(x['Read_Number']) for x in ins_info_main]
                    total_covs = median([int(x['Total_coverage']) for x in ins_info_main])
                    mq = median([float(x['MQ_median']) for x in ins_info_main])
                    # input()
                    muts = ','.join([str(x['Read_Muts']) for x in ins_info_main])
                    # print('ins_info_main', ins_info_main)
                    # print([str(x['Read_Muts']) for x in ins_info_main])
                    # print('muts', muts)
                    rnames = ','.join([str(x['Read_Names']) for x in ins_info_main])

                    left_seqs = [x['SBB'] for x in ins_info_main]      
                    right_seqs = [x['SAB'] for x in ins_info_main]            
                    

                    if ins_id not in uniq:
                        lr_file.write('\t'.join([str(ins_info_main[0]['Chrom']), # Chrom1
                                            str(min(map_poses))+','+str(max(poses)), # Pos1
                                            '+', # Strand1
                                            'D', # Junction_Side1
                                            str(ins_info_main[0]['Chrom']), # Chrom2
                                            str(round(mean(map_poses)))+','+str(round(mean(map_poses))+round(mean(map_lens))), # Pos2
                                            '+', # Strand2
                                            'D', # Jinction_Side2
                                            str(sum(rn)), # Read_Number
                                            str(max(total_covs - sum(rn), 0)), # Ref_Coverage
                                            str(mq), # MQ_median
                                            '99', # Read_Len1
                                            '99', # Read_Len2
                                            str(muts), # Read_Muts
                                            '0', # Inter_Joints
                                            '1', # Scarless_Joints
                                            '0', # NewSeq_Joints
                                            str(len(poses)), # Cigar_Num
                                            '0', # Inversion_Num
                                            '0', # Start_End_Locations
                                            '0', # Number_of_parts
                                            ','.join(left_seqs),
                                            ','.join(right_seqs),
                                            rnames # Read_names
                                            ]) + '\n')
                    for id in ins_ids:
                        uniq.append(id)  
        # input('STOP writing LRs')     

    def update_original_csv(self, output_file):
        print('Updating INS data...')
        logging.info('Updating INS data...')
        df = pd.read_csv(self.ins_file, delimiter='\t')
        filtered_df = df[~df.index.isin(self.tandem_duplications_ids)]
        filtered_df.to_csv(output_file, sep='\t', index=False)

    # def main(self):
    #     processor = InsertionProcessor(self.ins_file, self.lr_file, self.ref_genome, self.name, self.sam_file, self.threads, self.not_remove_trash)
    #     processor.read_insertions()
    #     fasta_file = self.name + '_insertions.fasta'
    #     processor.write_fasta(fasta_file)
    #     processor.map_with_minimap2(fasta_file)
    #     processor.analyze_mapping_results()
    #     processor.merge_tandem_duplications()
    #     processor.copy_file(self.lr_file, self.lr_file[:-4]+'_mapped_ins.csv')
    #     processor.extract_sequences_before_after()
    #     processor.update_large_rearrangements_file()
    #     updated_csv = self.ins_file[:-4]+'_mapped_ins.csv'
    #     processor.update_original_csv(updated_csv)

    #     print('Analysis of INS was done!')

# if __name__ == '__main__':
#     main()



# This script calls LRs in BAM file of ONT data
# 
import re
import sys
import argparse
import pysam
import math
from multiprocessing import Pool,Queue,Manager,Process,set_start_method
from multiprocessing.pool import ThreadPool
from threading import Thread
import statistics as stat
from tqdm import tqdm
import logging

class ONTLRCaller():

    def __init__(self,bamFile,log_path,
                 chrom=None,start=None,end=None,
                 threads=8,minVarLen=50,
                 minClipLen=100,distToJoinTrl=1000, workdir=''):
        if ((start==None or
             end==None) and 
            start!=end):
            print('ERROR (2)! If the start or the end for extracting reads is specified, another argument should also be defined.')
            print(start,end)
            exit(2)
        if bamFile[-2:] != 'gz':
            self.bam_file_name = bamFile.split('/')[-1][:-4]
        else:
            self.bam_file_name = bamFile.split('/')[-1][:-7]
        self.bamFile=bamFile
        self.log_path = log_path
        logging.basicConfig(level=logging.DEBUG, filename=self.log_path, filemode="a",
                            format="%(asctime)s %(funcName)s %(lineno)d %(message)s")
        self.workdir = workdir
        #self.outFile=outFile
        self.chrom=chrom
        self.start=start
        self.end=end
        self.threads=threads
        self.minVarLen=minVarLen
        self.minClipLen=minClipLen
        self.distToJoinTrl=distToJoinTrl
        # Stores coverage of genome regions splitted by minClipLen (e.g. 30 bp)
        # {chr1:{pos1:[cov1_primary,cov1_second],
        #        pos2:[cov1_primary,cov1_second]...}}
        self.genomeCoverage={}
        self.readToSecs={}
        #self.q=Queue()

    def readBamFile(self):
        global outFiles
        print('Reading BAM-file...')
        chromosomes=[self.chrom]       
        for chrom in chromosomes:
            if chrom not in self.genomeCoverage.keys():
                self.genomeCoverage[chrom]=[]
            # Collect all reads of one chromosome
            # {read_id:[minVarLen,minClipLen,read,sup_read1,sup_read2... ]}
            reads=[]
            # Stores secondary reads
            # Stores stat of read part intersections
            # [[start<end],[start==end],[start>end]]
            # For each variant, we store intersection length and its sequence
            # [intersection/distance length,sequence]
            # For start==end, we store +-1 bases
            bam=pysam.AlignmentFile(self.bamFile,threads=self.threads)
            if (self.start!=None and
                self.end!=None):
                regionForName='_'.join([str(chrom),
                                        str(self.start),
                                        str(self.end)])
            else:
                regionForName=str(chrom)
            for read in tqdm(bam.fetch(chrom,
                                  self.start,
                                  self.end)):
                # Calculate positions rounded to 30
                readStart=self.minVarLen*math.ceil(read.pos/self.minVarLen)
                readEnd=self.minVarLen*math.floor((read.pos+read.alen)/self.minVarLen)
                covValueNum=0
                # If read is secondary
                if read.is_secondary:
                    if read.query_name not in self.readToSecs.keys():
                        self.readToSecs[read.query_name]=[]
                    self.readToSecs[read.query_name].append([str(read.reference_name),
                                                        int(read.pos),
                                                        int(read.alen),
                                                        read.cigartuples,
                                                        read.is_reverse])
                    continue
                for pos in range(readStart,readEnd+self.minVarLen,self.minVarLen):
                    simplePos=int(math.floor(pos/self.minVarLen))
                    # If the position to add is more than the length of the list
                    if simplePos>=len(self.genomeCoverage[chrom]):
                        # We add all skipped elements
                        for i in range(len(self.genomeCoverage[chrom])-1,
                                       simplePos+1):
                            self.genomeCoverage[chrom].append(0)#[0,0]
                    self.genomeCoverage[chrom][simplePos]+=1 # [covValueNum]
                if (read.is_supplementary or
                    read.is_secondary):
                    continue
                # We take only primary mappings because they contain whole sequence
                # We save only necessary information about the read
                # [cigartuples,read_length,position,is_reverse,SA_tag]
                if read.has_tag('NM'):
                    nmTag=read.get_tag('NM')
                else:
                    nmTag=0
                reads.append([self.minVarLen,
                              self.minClipLen,
                              [],
                              [read.cigartuples,
                               read.infer_read_length(),
                               read.pos,
                               read.is_reverse,
                               read.query_sequence,
                               chrom,
                               read.alen,
                               read.query_name,
                               [read.mapping_quality],
                               nmTag]])
                if read.has_tag('SA'):
                    reads[-1][3].append(str(read.get_tag('SA')))
                else:
                    reads[-1][3].append([])
            if len(reads)==0:
                continue
            print('Analyzing reads for region '+regionForName+'...')
            logging.info('Analyzing reads for region '+regionForName+'...')
            # Create pool of processes
            p=ThreadPool(self.threads)
            # Write to output
            outFileName='.'.join([self.bam_file_name,
                                'junction_stat',regionForName,'xls'])
            outFileName = self.workdir + outFileName
            outFiles=[]
            outFileName1=outFileName[:outFileName.rfind('.')]+'.fusions.csv'
            outFiles.append(open(outFileName1,'w'))
            outFiles[-1].write('\t'.join(['Chrom1','Pos1','Strand1','Junction_Side1',
                                            'Chrom2','Pos2','Strand2','Junction_Side2',
                                            'MQ','Ref_Cov','Read_Len1','Read_Len2','Read_Muts',
                                            'Joint_Type','Joint_Length','Joint_Sequence',
                                            'First_Last_Part','Num_of_parts','Seq_before_BND','Seq_after_BND',
                                            'Read_Name'])+'\n')
            outFileName2=outFileName[:outFileName.rfind('.')]+'.insertions.csv'
            outFiles.append(open(outFileName2,'w'))
            outFiles[-1].write('\t'.join(['Chrom1','Pos1','Insertion_Length',
                                            'Sequence','Read_Number','Total_Cov',
                                            'MQ','Read_Muts1','Read_Name'])+'\n')
            outFileName3=outFileName[:outFileName.rfind('.')]+'.new_sequences.fa'
            outFileName4=outFileName[:outFileName.rfind('.')]+'.junctions.fa'
            outFiles.append(open(outFileName3,'w'))
            outFiles.append(open(outFileName4,'w'))
            for i,read in enumerate(reads):
                if read[3][7] in self.readToSecs.keys():
                    reads[i][2]=self.readToSecs[read[3][7]]
            results=[]
            m=Manager()
            self.thsize=0
            results=[]
            for res in tqdm(p.imap_unordered(self.callLRs,reads),
                            total=len(reads)):
                results.append(res)
            print('Writing fusions...')
            logging.info('Writing fusions...')
            for result in tqdm(results,
                                total=len(results)):
                t2=Thread(target=self.writeOutput,
                            args=(result,),daemon=True)
                t2.start()
                t2.join()
            for outFile in outFiles:
                outFile.close()

    def showPercWork(self,done,allWork):
        percDoneWork=round((done/allWork)*100,2)
        sys.stdout.write("\r"+str(percDoneWork)+"%")
        sys.stdout.flush()

    def makeAllRevCompl(self,seqsList):
        seqsListRC=[]
        for seq in seqsList:
            try:
                int(seq)
                seqsListRC.append(seq)
            except ValueError:
                seqsListRC.append(self.revComplement(seq))
        return(seqsListRC)

    def revComplement(self,seq):
        nucToRC={'A':'T','T':'A','G':'C','C':'G',
                 'a':'t','t':'a','g':'c','c':'g'}
        seqRC=[]
        for n in seq:
            seqRC.append(nucToRC[n])
        return(''.join(seqRC))        

    def getInDelsFromRead(self,cigar,chrom,curRefPos,
                          readSeq,mapQual,NM,
                          name,minVarLen):
        deletions=[]
        insertions=[]
        ciTypeStrToInt={'M':0,'I':1,'D':2,
                        'N':3,'S':4,'H':5,
                        'P':6,'X':7,'=':9}                    
        # If cigar is provided as string, convert it to tuple
        if type(cigar)==str:
            cigarTuples=[]
            cigarPat=re.compile(r'(\d+)([MIDNSHPX=]+)')
            for ciLen,ciType in cigarPat.findall(cigar):
                cigarTuples.append((ciTypeStrToInt[ciType],
                                    int(ciLen)))
        else:
            cigarTuples=cigar
        # Extract DELs and INSs from CIGAR
        curReadPos=0
        alignedLenPart1=0
        for ciType,ciLen in cigarTuples:
            if (ciType in [1,2] and
                ciLen>=minVarLen):
                # If it is insertion
                if ciType==1:
                    insertedSeq=readSeq[curReadPos:curReadPos+ciLen]
                    if (len(insertions)>0 and
                        abs(insertions[-1][1]-curRefPos)<200):
                        insertions[-1][2]+=ciLen
                        insertions[-1][5]+=insertedSeq
                    else:
                        insertions.append([chrom,
                                           curRefPos,
                                           ciLen,
                                           mapQual,NM,
                                           insertedSeq,
                                           name])
                elif ciType==2:
                    deletions.append([chrom,
                                      curRefPos,
                                      curRefPos+ciLen-1,
                                      mapQual,NM,
                                      curReadPos,
                                      alignedLenPart1])
            if ciType in [0,2,3]:
                curRefPos+=ciLen
                alignedLenPart1+=ciLen
            if ciType in [0,1,4]:
                curReadPos+=ciLen
        return(deletions,
               insertions)

    # The following function calls large rearrangements from read object of pysam
    def callLRs(self,readInfo):
        self.thsize+=1
        minVarLen=readInfo[0]
        minClipLen=readInfo[1]
        # Contains supplementary alignment positions in the read
        # [[16,181],[180,1152]...]
        readSupPoses=[]
        # We take the first aligment met and check for some fusions
        # [cigartuples,read_length,position,
        #  is_reverse,seq,chrom,alen,name,[MQ],NM,SA_tag]
        mainReadPart=readInfo[3]
        secAligns=readInfo[2]
        # Get main read part strand
        if mainReadPart[3]:
            mainReadPartStrand='-'
        else:
            mainReadPartStrand='+'
        # Deletions are stored as [start,end]
        deletions=[]
        # Insertions are stored as [position,length,seq]
        insertions=[]
        # Stores CIGAR regular expression
        cigarPat=re.compile(r'(\d+)([MIDNSHPX=]+)')
        if mainReadPart[0]==None:
            return([],mainReadPart,
                   deletions,insertions)
        # Extract DELs and INSs from CIGAR
        inDels=self.getInDelsFromRead(mainReadPart[0],
                                 mainReadPart[5],
                                 mainReadPart[2]+1,
                                 mainReadPart[4],
                                 mainReadPart[8],
                                 mainReadPart[9], # NM
                                 mainReadPart[7],
                                 minVarLen)
        deletions.extend(inDels[0])
        insertions.extend(inDels[1])
        # If the left or the right part of the cigar contains clipped bases
        if (((mainReadPart[0][0][0] in [4,5] and
              # and it is at least of minimal accepted length
              mainReadPart[0][0][1]>=minClipLen) or
             (mainReadPart[0][-1][0] in [4,5] and
              # and it is at least of minimal accepted length
              mainReadPart[0][-1][1]>=minClipLen)) and
            # One more IMPORTANT condition is the presence of supplementary alignments for the read
            # IMPORTANT! We do not consider bases clipped at read ends as LRs
            len(mainReadPart[10])>0):
            # Stores left or right side by which this part is connected to preavious and next parts, respectively
            sidesForConn=['N','N']
            # Determine coordinates of the main alignment in the read
            # Start is a number of clipped bases +1
            if mainReadPart[0][0][0] in [4,5]:
                start=mainReadPart[0][0][1]+1
                # if it is on '-' strand
                if mainReadPart[3]:
                    sidesForConn[0]='L'
                else:
                    sidesForConn[0]='L'
            else:
                start=1
                
            # End is a start+aligned_length-1
            if mainReadPart[0][-1][0] in [4,5]:
                end=mainReadPart[1]-mainReadPart[0][-1][1]
                # if it is on '-' strand
                if mainReadPart[3]:
                    sidesForConn[1]='R'
                else:
                    sidesForConn[1]='R'
            else:
                end=mainReadPart[1]
            # Add this part to read parts
            # startInReadSeq,endInReadSeq,chrom,pos,alen,NM,strand
            readSupPoses.append([start,end,
                                 mainReadPart[5],
                                 mainReadPart[2]+1,
                                 mainReadPart[6],
                                 int(mainReadPart[9]),
                                 mainReadPartStrand,
                                 sidesForConn])
            # Then go through all supplementary alignments and determine their coordinates in the read
            for readSup in mainReadPart[10].split(';')[:-1]:
                # Get supplementary info
                supChrom,supCoord,supStrand,supCigar,supMQ,supNM=readSup.split(',')
                # We check if some secondary mapping with better alignment has been saved
                mainReadPart[8].append(int(supMQ))
                # Stores left or right side by which this part is connected to previous and next parts, respectively
                sidesForConn=['N','N']
                # If strand of supplementary is the same as in the main part
                if mainReadPartStrand==supStrand:
                    # Then its CIGAR coordinates correspond main read part
                    cigarMatches=cigarPat.findall(supCigar)
                    # We determine this supplement alignment position in the read
                    # If its 1st CIGAR part is S
                    if cigarMatches[0][1]=='S':
                        # Then the start is the number of clipped bases +1
                        start=int(cigarMatches[0][0])+1
                        # If supplementary is on '+' strand
                        if supStrand=='+':
                            if mainReadPartStrand=='+':
                                sidesForConn[0]='L'
                            else:
                                sidesForConn[1]='L'
                        else:
                            if mainReadPartStrand=='+':
                                sidesForConn[1]='L'
                            else:
                                sidesForConn[0]='L'
                    else:
                        # Then the start is 1
                        start=1
                    # If the last CIGAR part is S
                    if cigarMatches[-1][1]=='S':
                        # Then the end is a (read_length - the_number_of_clipped_bases) 
                        end=mainReadPart[1]-int(cigarMatches[-1][0])
                        # If supplementary is on '+' strand
                        if supStrand=='+':
                            if mainReadPartStrand=='+':
                                sidesForConn[1]='R'
                            else:
                                sidesForConn[0]='R'
                        else:
                            if mainReadPartStrand=='+':
                                sidesForConn[0]='R'
                            else:
                                sidesForConn[1]='R'
                    else:
                        # Then end is a read length
                        end=mainReadPart[1]
                    # Calculate aligned part length
                    alen=0
                    for cigarPart in cigarMatches:
                        if cigarPart[1] in ['M','D']:
                            alen+=int(cigarPart[0])
                    if alen>=minClipLen:
                        for secReadPart in secAligns:
                            distToMain=abs(secReadPart[1]-mainReadPart[2])
                            distToSup=abs(secReadPart[1]-int(supCoord))
                            mainToSupDist=abs(int(supCoord)-mainReadPart[2])
                            # Compare values of the length of aligned parts for SA
                            if secReadPart[2]/alen>=0.8:
                                # Compare chromosomes
                                if (secReadPart[0]==mainReadPart[5] and
                                    (supChrom!=mainReadPart[5] or
                                     (distToMain+1000<min(10000,mainToSupDist) and
                                      distToSup>distToMain))):
                                    return([],mainReadPart,
                                           deletions,insertions)
                                    # If chromosome is the same like in the primary alignment and it's closer to it than current sup_align
                                    # Then, now we need to check that current secondary alignment refers to the same part of read like sup_align
                            # Compare values of the length of aligned parts for main mapping
                            if secReadPart[2]/mainReadPart[6]>=0.8:
                                # Compare chromosomes
                                if (secReadPart[0]==supChrom and
                                    (supChrom!=mainReadPart[5] or
                                     distToSup+1000<min(10000,mainToSupDist))):
                                    return([],mainReadPart,
                                           deletions,insertions)
                                    # If chromosome is the same like in the supplementary alignment and it's closer to it than current primary_al
                                    # Then, now we need to check that current secondary alignment refers to the same part of read like sup_align
                        # Add this read part to read parts
                        # startInReadSeq,endInReadSeq,chrom,pos,alen,strand
                        readSupPoses.append([start,end,
                                             supChrom,int(supCoord),
                                             alen,int(supNM),supStrand,
                                             sidesForConn])
                # If strand of supplementary is the opposite to the main part strand
                else:
                    # Then its CIGAR coordinates correspond main read part but reversed
                    cigarMatches=cigarPat.findall(supCigar)
                    # We determine this supplement alignment position in the read
                    # If its last CIGAR part is S
                    if cigarMatches[-1][1]=='S':
                        # Then the start is the number of clipped bases +1
                        start=int(cigarMatches[-1][0])+1
                        # If supplementary is on '+' strand
                        if supStrand=='+':
                            if mainReadPartStrand=='+':
                                sidesForConn[1]='R'
                            else:
                                sidesForConn[0]='R'
                        else:
                            if mainReadPartStrand=='+':
                                sidesForConn[0]='R'
                            else:
                                sidesForConn[1]='R'
                    else:
                        # Then the start is 1
                        start=1
                    # If the 1st CIGAR part is S
                    if cigarMatches[0][1]=='S':
                        # Then the end is a (read_length - the_number_of_clipped_bases) 
                        end=mainReadPart[1]-int(cigarMatches[0][0])
                        # If supplementary is on '+' strand
                        if supStrand=='+':
                            if mainReadPartStrand=='+':
                                sidesForConn[0]='L'
                            else:
                                sidesForConn[1]='L'
                        else:
                            if mainReadPartStrand=='+':
                                sidesForConn[1]='L'
                            else:
                                sidesForConn[0]='L'
                    else:
                        # Then end is a read length
                        end=mainReadPart[1]
                    # Calculate aligned part length
                    alen=0
                    for cigarPart in cigarMatches:
                        if cigarPart[1] in ['M','D']:
                            alen+=int(cigarPart[0])
                    if alen>=minClipLen:
                        for secReadPart in secAligns:
                            distToMain=abs(secReadPart[1]-mainReadPart[2])
                            distToSup=abs(secReadPart[1]-int(supCoord))
                            mainToSupDist=abs(int(supCoord)-mainReadPart[2])
                            # Compare values of the length of aligned parts for SA
                            if secReadPart[2]/alen>=0.8:
                                # Compare chromosomes
                                if (secReadPart[0]==mainReadPart[5] and
                                    (supChrom!=mainReadPart[5] or
                                     (distToMain+1000<min(10000,mainToSupDist) and
                                      distToSup>distToMain))):
                                    return([],mainReadPart,
                                           deletions,insertions)
                                    # If chromosome is the same like in the primary alignment and it's closer to it than current sup_align
                                    # Then, now we need to check that current secondary alignment refers to the same part of read like sup_align
                            # Compare values of the length of aligned parts for main mapping
                            if secReadPart[2]/mainReadPart[6]>=0.8:
                                # Compare chromosomes
                                if (secReadPart[0]==supChrom and
                                    (supChrom!=mainReadPart[5] or
                                     distToSup+1000<min(10000,mainToSupDist))):
                                    return([],mainReadPart,
                                           deletions,insertions)
                                    # If chromosome is the same like in the primary alignment and it's closer to it than current sup_align
                                    # Then, now we need to check that current secondary alignment refers to the same part of read like sup_align
                        # Add this read part to read parts
                        # startInReadSeq,endInReadSeq,chrom,pos,alen,strand
                        readSupPoses.append([start,end,
                                             supChrom,int(supCoord),
                                             alen,int(supNM),supStrand,
                                             sidesForConn])
        return(readSupPoses,mainReadPart,
               deletions,insertions)

    def countHomoNucNum(self,seq,nuc='T',
                        minLen=4):
        p=re.compile(nuc+'+')
        n=0
        sumLen=0
        for m in p.findall(seq):
            if len(m)>=minLen:
                n+=1
                sumLen+=len(m)
        return(n,sumLen)

    def countDifNucHomoTracts(self,seq):
        nucToHomoSumLen={'A':0,'T':0,'G':0,'C':0}
        nucToHomoTractNum={'A':0,'T':0,'G':0,'C':0}
        for nuc in nucToHomoSumLen.keys():
            n,sumLen=self.countHomoNucNum(seq,nuc=nuc)
            nucToHomoSumLen[nuc]=sumLen*100/len(seq)
            nucToHomoTractNum[nuc]=n #*100/len(seq)
        return(max(nucToHomoSumLen.values()),sum(nucToHomoTractNum.values()))

    def writeOutput(self,out):
        global outFiles
        self.thsize+=1
        fusions=[]
        try:
            readParts,(cigarTuples,readLen,readPos,readReverse,readSeq,readChrom,readAlignedLen,readName,mapQual,mapNM,readSups),deletions,insertions=out#out.get()
        except ValueError:
            print('ERROR',out)
            exit()
        # Process saved read parts and their sequences
        # First, sort read parts
        readParts.sort()
        fusion=None
        # Stores number of BND saved
        bndSaved=0
        homoTractSumValue=0
        # Go through all read parts and compare positions of the neighboring parts
        for i,readPart in enumerate(readParts[1:]):
            fusion=[]
            # Add info from readParts
            # startInReadSeq,endInReadSeq,chrom,pos,alen,strand
            # Add chrom
            # 'Chrom1'
            fusion.append(readParts[i][2])
            # 'Pos1'
            # For the previous part, we take end genome position regardless of its strand
            if readParts[i][7][1]=='R':
                fusion.append(readParts[i][3]+readParts[i][4]-1)
            elif readParts[i][7][1]=='L':
                fusion.append(readParts[i][3])
            else:
                break
            # 'Strand1','Junction_Side1'
            fusion.append(readParts[i][6])
            fusion.append(readParts[i][7][1])
            # 'Chrom2'
            fusion.append(readPart[2])
            # 'Pos2'
            if readPart[7][0]=='R':
                fusion.append(readPart[3]+readPart[4]-1)
            elif readPart[7][0]=='L':
                fusion.append(readPart[3])
            else:
                break
            # 'Strand2','Junction_Side2'
            fusion.append(readPart[6])
            fusion.append(readPart[7][0])
            # MQ
            fusion.append(stat.median(mapQual))
            # 'Read_Len1','Read_Len2'
            fusion.append(readParts[i][4])
            fusion.append(readPart[4])
            # NM
            errPerc=[str(int(round(int(readParts[i][5])*100/int(readParts[i][4]),0))),
                        str(int(round(int(readPart[5])*100/int(readPart[4]),0)))]
            fusion.append(','.join(errPerc))
            # Get junction sequence
            junctionSeq=readSeq[readParts[i][0]:readPart[1]]
            junctionPos1=readParts[i][1]-readParts[i][0]
            junctionPos2=readPart[0]-readParts[i][0]
            homoTractSumValue=0
            # Choose variant how we will write junction seq
            # If there is intersection or scarless junction
            #         -------<----->-----    ---||---
            # Or sequence between them is short (less 50 bp)
            #    --------->~~~~~~~~~~~~~<-----------------
            if (readPart[0]<=readParts[i][1]+1 or
                readPart[0]-readParts[i][1]<8):
                outFiles[3].write('_'.join(['>'+readName,
                                            str(fusion[0]),
                                            str(fusion[1]),
                                            str(fusion[4]),
                                            str(fusion[5]),
                                            'not_long_dist',
                                            str(junctionPos1),
                                            str(junctionPos2)])+'\n')
                outFiles[3].write(junctionSeq+'\n')
            else:
                # Else we write junction of left part with distseq
                outFiles[3].write('_'.join(['>'+readName,
                                            'long_dist1',
                                            str(fusion[0]),
                                            str(fusion[1]),
                                            str(fusion[4]),
                                            str(fusion[5]),
                                            str(junctionPos1),
                                            str(junctionPos2)])+'\n')
                outFiles[3].write(junctionSeq+'\n')
            # Then we write whole left part and whole right part
            # If start of the next less than end of the previous (intersection)
            if readPart[0]<=readParts[i][1]:
                interLen=readParts[i][1]-readPart[0]+1
                seq=readSeq[readPart[0]-1:readParts[i][1]+1-1]
                fusion.extend(['inter',
                                interLen,
                                seq])
                
            # If start==end+1
            elif readPart[0]-readParts[i][1]==1:
                interLen=1
                seq=readSeq[readPart[0]-1-1:readPart[0]+1-1]
                fusion.extend(['scarless',
                                interLen,
                                seq])
            # If start of the next more than end of the previous (distance)
            elif readPart[0]>readParts[i][1]:
                interLen=readPart[0]-readParts[i][1]-1
                seq=readSeq[readParts[i][1]+1-1:readPart[0]-1]
                homoTractSumLen,homoTractSumValue=self.countDifNucHomoTracts(seq)
                outFiles[2].write('_'.join(['>'+readName,
                                            str(fusion[0]),
                                            str(fusion[1]),
                                            str(fusion[4]),
                                            str(fusion[5]),
                                            str(junctionPos1),
                                            str(junctionPos2),
                                            str(readParts[i][1]+1),
                                            str(readPart[0]-1),
                                            str(homoTractSumValue)])+'\n')
                outFiles[2].write(seq+'\n')
                if interLen>6:
                    seq=''
                fusion.extend(['newseq',
                                interLen,
                                seq])
            # Add seqBeforeBND and seqAfterBND
            seqBefore=readSeq[max(0,readParts[i][1]-30):readParts[i][1]]
            seqAfter=readSeq[readPart[0]-1:readPart[0]-1+30]
            # Add First_Last_Part, Number_of_parts, seq_before_BND, seq_after_BND
            if i==0:
                firstLastPart=1
            elif i==len(readParts)-2:
                firstLastPart=2
            else:
                firstLastPart=0
            fusions.append(fusion+[homoTractSumValue,
                                    len(readParts),
                                    seqBefore,
                                    seqAfter,
                                    readName])
        # If number of saved BNDs is more than 2
        # then we need to try determine the 2nd coordinate of the LGR
        if len(fusions)>=2:
            fusionsJoint=[]
            for i in range(1,len(fusions)):
                # Go through all fusion boundaries and compare their
                # strands (+/-), positions (>/<) and side of junction (L/R)
                pos1,pos2,pos3,pos4=[fusions[i-1][1],
                                        fusions[i-1][5],
                                        fusions[i][1],
                                        fusions[i][5]]
                # Joint strand+size lists
                ss1,ss2,ss3,ss4=[fusions[i-1][2]+fusions[i-1][3],
                                    fusions[i-1][6]+fusions[i-1][7],
                                    fusions[i][2]+fusions[i][3],
                                    fusions[i][6]+fusions[i][7]]
                # Determine end of the 1st fragment
                if ss1[1]=='L':
                    fragEnd1=pos1+fusions[i-1][9]
                elif ss1[1]=='R':
                    fragEnd1=pos1-fusions[i-1][9]
                # Determine end of the 1st fragment
                if ss2[1]=='L':
                    fragEnd2=pos4+fusions[i-1][10]
                elif ss2[1]=='R':
                    fragEnd2=pos4-fusions[i-1][10]
                ssOrder=[ss1,ss2,ss3,ss4]
                trlSides=[['+R','+L','+R','+L'],
                            ['+R','-R','-L','+L'],
                            ['-L','+L','+R','-R'],
                            ['-L','-R','-L','-R']]
                tdSides=[['+R','+L','+R','+L'],
                            ['-L','-R','-L','-R']]
                invSides=[['-L','+L','+R','-R'],
                            ['+R','-R','-L','+L']]
                invTdSides=[['-L','+L','+R','-R'],
                            ['+R','-R','-L','+L'],
                            ['+R','-R','-L','-R'],
                            #['+R','+L','+R','-R'],
                            ['-L','+L','+R','+L']]
                # Get sequences before and after BND
                posList=[pos1,pos2,pos3,pos4]
                errPerc=','.join([fusions[i-1][11],
                                    fusions[i][11]])
                jointType=[fusions[i-1][12],
                            fusions[i][12]]
                jointLen=[str(fusions[i-1][13]),
                            str(fusions[i][13])]
                jointSeq=[str(fusions[i-1][14]),
                            str(fusions[i][14])]
                firstLastPart=[str(fusions[i-1][15]),
                                str(fusions[i][15])]
                seqsNearBND=[str(fusions[i-1][17]),
                                str(fusions[i][17]),
                                str(fusions[i-1][18]),
                                str(fusions[i][18])]
                if ss1[1]=='L':
                    if ss1[0]=='-':
                        posList=[str(pos4),str(pos1),str(pos3),str(pos2)]
                    else:
                        posList=[str(pos4),str(pos1),str(pos3),str(pos2)]
                    jointType=jointType[::-1]
                    jointLen=jointLen[::-1]
                    jointSeq=jointSeq[::-1]
                    firstLastPart=firstLastPart[::-1]
                    seqsBeforeBNDs=seqsNearBND[::-1][:2]
                    seqsAfterBNDs=seqsNearBND[::-1][2:]
                else:
                    if ss1[0]=='-':
                        posList=[str(pos1),str(pos4),str(pos2),str(pos3)]
                    else:
                        posList=[str(pos1),str(pos4),str(pos2),str(pos3)]
                    seqsBeforeBNDs=seqsNearBND[:2]
                    seqsAfterBNDs=seqsNearBND[2:]
                if ss1[0]=='-':
                    jointSeq=self.makeAllRevCompl(jointSeq)
                    seqsBeforeBNDs=self.makeAllRevCompl(seqsBeforeBNDs)
                    seqsAfterBNDs=self.makeAllRevCompl(seqsAfterBNDs)
                # Translocation
                twoJointLR=False
                # Check order of strands and sides
                if ((ssOrder in trlSides or
                        ssOrder[::-1] in trlSides) and
                    # Check if position of the 2nd part is out of the 1st fragment location
                    fusions[i-1][0]==fusions[i][4] and # chrom of 1st and 4th positions are the same
                    abs(pos4-pos1)<=self.distToJoinTrl and # distance between two positions flanking position into which TRL was inserted should be not so large
                    (fusions[i-1][0]!=fusions[i-1][4] or # chrom for 2nd and 3rd position are another
                        (pos2<min(pos1,fragEnd1,pos4,fragEnd2) and pos3<min(pos1,fragEnd1,pos4,fragEnd2) or # or those positions are out of range of both fragments
                                pos2>max(pos1,fragEnd1,pos4,fragEnd2) and pos3>max(pos1,fragEnd1,pos4,fragEnd2)))):
                    # We need to write it in the new form
                    # chrom1 - chromosome of position of genome where new fragment was added
                    # pos1 - position of genome where new fragment was added
                    # strand1 - it is always +, because we don't need to write it in opposite direction
                    # side1 - it was L or R, for translocations it will be T
                    # chrom2 - chromosome of the position of genome from which fragment was transfered
                    # pos2,pos3 - positions of genome from which fragment was transfered
                    # mq - mean mapping quality
                    # fraglen1 - length of one fragment around site of new fragment insertion
                    # fraglen2 - length of the 2nd fragment around site of new fragment insertion
                    # jointType- type of joinment, listed with separation by comma for two junctions
                    # jointLen - length of the joinment, listed with separation by comma for two junctions
                    # jointSeq - sequence of the joinment, listed with separation by comma for two junctions
                    # firstLastPart - is it first or the last part of the read (1 - 1st, 2 - last, 0 - other)
                    # partsNum - number of read parts
                    # seqBeforeBND - sequence before juntions, listed with separation by comma for two junctions
                    # seqAfterBND - sequence after juntions, listed with separation by comma for two junctions
                    # readName - Read_Name
                    mq=str(round((fusions[i-1][8]+fusions[i][8])/2,0))
                    fusionsJoint.append([fusions[i-1][0],','.join(posList[:2]),'+','T',
                                            fusions[i-1][4],','.join(posList[2:]),
                                            '+','T',mq,
                                            fusions[i-1][9], # fraglen1
                                            fusions[i-1][10], # fraglen2
                                            errPerc, # errPercents
                                            ','.join(jointType), # jointType
                                            ','.join(jointLen), # jointLen
                                            ','.join(jointSeq), # jointSeq
                                            # number was deleted from fusions
                                            ','.join(firstLastPart), # firstLastPart
                                            fusions[i-1][16], # partsNum
                                            ','.join(seqsBeforeBNDs), # seqBeforeBND
                                            ','.join(seqsAfterBNDs), # seqAfterBND
                                            fusions[i][19]])
                    ############################################################
                    ############### Check that some sequence was duplicated around TRL, e.g. chr1:1929454,1929364
                    # chr1:8,912,128-8,912,409
                    # chr1    8912269,8912173 +       T       chr15   43014532,43015654       +       
                    # T       49.0    6       216     1123    inter,inter     94,608
                    ############################################################
                    twoJointLR=True
                elif ((ssOrder in tdSides or
                        ssOrder[::-1] in tdSides) and
                        fusions[i-1][0]==fusions[i-1][4]==fusions[i][0]==fusions[i][4] and
                        abs(pos1-pos3)<100 and
                        abs(pos2-pos4)<100):
                    mq=str(round((fusions[i-1][8]+fusions[i][8])/2,0))
                    posList=[str(pos1),str(pos2),str(pos3),str(pos4)]
                    if ss1[1]=='L':
                        posList=posList[::-1]
                    # Calculate distance of the third fragment to the end of the repeat
                    # If it is far from it, we should subtract this distanca from the length of the TD
                    lenBias=max(0,int(posList[0])-int(fusions[i-1][9])-int(posList[1]),
                                int(posList[2])-(int(posList[3])+int(fusions[i][10])))
                    # If such fragment did not reach the end the repeat we need take only the small repeat length
                    # ++++++-------------++++++ - ref
                    # ++++++-------------++++++-------------++++++-------------++++++ - TD
                    #           ---------++++++ - third fragment
                    if lenBias>max([fusions[i-1][13],
                                    fusions[i][13]]):
                        lenBias=max([fusions[i-1][13],
                                        fusions[i][13]])
                    # Check, if the previous fusionJoint was also TD
                    # Then, it means that it is the extension of the previous TD
                    if (len(fusionsJoint)>0 and
                        fusionsJoint[-1][3][-1]=='D'):
                        # And then we work with the previous fusionJoint, but not this one
                        # Check, if this extension of previous TD has been already done
                        if len(fusionsJoint[-1][3])>1:
                            tdSize=int(fusionsJoint[-1][3].replace('D',''))
                            fusionsJoint[-1][3]=str(tdSize+1)+'D'
                            fusionsJoint[-1][7]=str(tdSize+1)+'D'
                        else:
                            fusionsJoint[-1][3]='2D'
                            fusionsJoint[-1][7]='2D'
                        fusionsJoint[-1][10]=fusions[i][10]
                        prevLenBias=int(fusionsJoint[-1][9].split(',')[1])
                        if prevLenBias<lenBias:
                            fusionsJoint[-1][9]=','.join([fusionsJoint[-1][9].split(',')[0],
                                                            str(lenBias)])
                        fusionsJoint[-1][11]+=','+errPerc
                        fusionsJoint[-1][12]+=','+','.join(jointType)
                        fusionsJoint[-1][13]+=','+','.join(jointLen)
                        fusionsJoint[-1][14]+=','+','.join(jointSeq)
                        fusionsJoint[-1][15]+=','+','.join(firstLastPart)
                        fusionsJoint[-1][17]+=','+','.join(seqsBeforeBNDs)
                        fusionsJoint[-1][18]+=','+','.join(seqsAfterBNDs)
                    else:
                        fusionsJoint.append([fusions[i-1][0],','.join(posList[:2]),'+','D',
                                                fusions[i-1][4],','.join(posList[2:]),'+','D',mq,
                                                ','.join([str(fusions[i-1][9]),
                                                        str(lenBias)]), # fraglen1 of fusion1
                                                fusions[i][10], # fraglen2 of fusion2
                                                errPerc, # errPercents
                                                ','.join(jointType), # jointType
                                                ','.join(jointLen), # jointLen
                                                ','.join(jointSeq), # jointSeq
                                                # number was deleted from fusions
                                                ','.join(firstLastPart), # firstLastPart
                                                fusions[i-1][16], # partsNum
                                                ','.join(seqsBeforeBNDs), # seqBeforeBND
                                                ','.join(seqsAfterBNDs), # seqAfterBND
                                                fusions[i][19]])
                    twoJointLR=True
                elif ((ssOrder in invSides or
                        ssOrder[::-1] in invSides) and
                        fusions[i-1][0]==fusions[i-1][4]==fusions[i][0]==fusions[i][4] and
                        pos1<=pos2<=pos4 and
                        pos1<=pos3<=pos4 and
                        abs(pos1-pos2)>=self.minVarLen and
                        abs(pos3-pos4)>=self.minVarLen):
                    mq=str(round((fusions[i-1][8]+fusions[i][8])/2,0))
                    fusionsJoint.append([fusions[i-1][0],','.join([str(min(pos1,pos4)),
                                                                    str(min(pos2,pos3))]),
                                            '+','I',
                                            fusions[i-1][4],','.join([str(max(pos2,pos3)),
                                                                    str(max(pos1,pos4))]),
                                            '+','I',mq,
                                            fusions[i-1][9], # fraglen1
                                            fusions[i-1][10], # fraglen2
                                            errPerc, # errPercents
                                            ','.join(jointType), # jointType
                                            ','.join(jointLen), # jointLen
                                            ','.join(jointSeq), # jointSeq
                                            # number was deleted from fusions
                                            ','.join(firstLastPart), # firstLastPart
                                            fusions[i-1][16], # partsNum
                                            ','.join(seqsBeforeBNDs), # seqBeforeBND
                                            ','.join(seqsAfterBNDs), # seqAfterBND
                                            fusions[i][19]])
                    # Add detection of deletions near the inversion
                    # If the distance between two fragments is more than the LR minimal length
                    # We save it as deletion
                    if abs(pos1-pos3)>=self.minVarLen:
                        fusionsJoint.append([fusions[i-1][0],str(min(pos1,pos3)),'+','R',
                                                fusions[i-1][4],str(max(pos2,pos3)),'+','L',mq,
                                                fusions[i-1][9], # fraglen1
                                                fusions[i-1][10], # fraglen2
                                                errPerc, # errPercents
                                                'inversion',
                                                0,'',
                                                # number was deleted from fusions
                                                0,
                                                fusions[i-1][15], # partsNum
                                                '', # seqBeforeBND
                                                '', # seqAfterBND
                                                fusions[i][18]])
                    if abs(pos2-pos4)>=self.minVarLen:
                        fusionsJoint.append([fusions[i-1][0],str(min(pos2,pos4)),'+','R',
                                                fusions[i-1][4],str(max(pos2,pos4)),'+','L',mq,
                                                fusions[i][9], # fraglen1
                                                fusions[i][10], # fraglen2
                                                errPerc, # errPercents
                                                'inversion',
                                                0,'',
                                                # number was deleted from fusions
                                                0,
                                                fusions[i-1][16], # partsNum
                                                '', # seqBeforeBND
                                                '', # seqAfterBND
                                                fusions[i][19]])
                    twoJointLR=True
                elif ((ssOrder in invTdSides or
                        ssOrder[::-1] in invTdSides) and
                        fusions[i-1][0]==fusions[i-1][4]==fusions[i][0]==fusions[i][4] and
                        abs(pos1-pos4)<100 and
                        (abs(pos1-pos2)<100 or
                        abs(pos3-pos4)<100)):
                    mq=str(round((fusions[i-1][8]+fusions[i][8])/2,0))
                    fusionsJoint.append([fusions[i-1][0],','.join(posList[:2]),
                                            '+','V',
                                            fusions[i-1][4],','.join(posList[2:]),
                                            '+','V',mq,
                                            fusions[i-1][9], # fraglen1
                                            fusions[i-1][10], # fraglen2
                                            errPerc, # errPercents
                                            ','.join(jointType), # jointType
                                            ','.join(jointLen), # jointLen
                                            ','.join(jointSeq), # jointSeq
                                            # number was deleted from fusions
                                            ','.join(firstLastPart), # firstLastPart
                                            fusions[i-1][16], # partsNum
                                            ','.join(seqsBeforeBNDs), # seqBeforeBND
                                            ','.join(seqsAfterBNDs), # seqAfterBND
                                            fusions[i][19]])
                    twoJointLR=True
                else:
                    fusionsJoint.append(fusions[i-1])
                    if i==len(fusions)-1:
                        fusionsJoint.append(fusions[i])
            if twoJointLR==True:
                fusions=fusionsJoint[:]
        # Write deletions and insertions
        for deletion in deletions:
            errPerc=int(round(deletion[4]*100/readAlignedLen,0))
            fusion=[deletion[0],str(deletion[1]),'+','R',
                    deletion[0],str(deletion[2]),'+','L',
                    stat.median(deletion[3]),
                    deletion[6],readAlignedLen-deletion[6]-(deletion[2]-deletion[1]+1),errPerc,'cigar',0,'']
            fusions.append(fusion+[0,
                                    len(readParts),
                                    readSeq[max(0,deletion[5]-30):deletion[5]],
                                    readSeq[deletion[5]:deletion[5]+30],
                                    readName])
            anyFusionsFound=True
        for fusion in fusions:
            # [chr1,pos1,strand1,side1,
            #  chr2,pos2,strand2,side2,
            #  MQ,refCov,readLen1,readLen2,
            #  type(intersection,equal, or distance),
            #  length,sequence,number,
            #  firstLastPart,partsNum,
            #  seqBeforeBND,seqAfterBND]
            out=[]
            for fusionInfoNum in range(8):
                out.append(str(fusion[fusionInfoNum])) 

            ###########################################################################################
            ####### TO DO: Analyze coverage statistics to get additional characteristics for DELs #####
            ###########################################################################################
            # MQ
            out.append(str(fusion[8]))
            pos1=int(str(fusion[1]).split(',')[0])
            if (fusion[3]=='T' or
                'D' in fusion[3] or
                fusion[3]=='V'):
                pos2=pos1
            elif fusion[3]=='I':
                pos2=int(str(fusion[5]).split(',')[1])
            else:
                pos2=int(fusion[5])
            # Get reference coverage for the left coordinate
            if (fusion[3]=='R' or
                fusion[3]=='T' or
                'D' in fusion[3] or
                fusion[3]=='I' or
                fusion[3]=='V'):
                leftCoord=self.minVarLen*math.ceil(pos1/self.minVarLen)-self.minVarLen
            elif fusion[3]=='L':
                leftCoord=self.minVarLen*math.floor(pos1/self.minVarLen)+self.minVarLen
            # If fusion's chromosome is not the same as the current input
            leftCoordSimple=int(math.floor(leftCoord/self.minVarLen))
            if (fusion[0] in self.genomeCoverage.keys() and
                leftCoordSimple<len(self.genomeCoverage[fusion[0]])):
                leftCov=self.genomeCoverage[fusion[0]][leftCoordSimple]#[0]
            else:
                leftCov=0
            # Get reference coverage for the right coordinate
            if fusion[7]=='R':
                rightCoord=self.minVarLen*math.ceil(pos2/self.minVarLen)-self.minVarLen
            elif (fusion[7]=='L' or
                    fusion[7]=='T' or
                    'D' in fusion[7] or
                    fusion[7]=='I' or
                    fusion[7]=='V'):
                rightCoord=self.minVarLen*math.floor(pos2/self.minVarLen)+self.minVarLen
            # If fusion's chromosome is not the same as the current input
            rightCoordSimple=int(math.floor(rightCoord/self.minVarLen))
            if (fusion[4] in self.genomeCoverage.keys() and
                rightCoordSimple<len(self.genomeCoverage[fusion[4]])):
                rightCov=self.genomeCoverage[fusion[4]][rightCoordSimple]#[0]
            else:
                rightCov=0
            # For intrachromosomal we get maximal
            if fusion[0]==fusion[4]:
                out.append(str(max(leftCov,rightCov)))
            # For interchromosomal, we get sum
            else:
                out.append(str(max([leftCov,rightCov])))
            try:
                out.append(str(fusion[9]))
            except TypeError:
                print('ERROR!',fusion)
                print(fusion[9])
                print(fusion[10])
                exit()
            for fusionInfoNum in range(10,20):
                out.append(str(fusion[fusionInfoNum]))
            outFiles[0].write('\t'.join(out)+'\n')
        # chrom,curRefPos,ciLen,mapQual,insertedSeq,readName
        for insertion in insertions:
            out=[]
            out.extend([insertion[0],
                        str(insertion[1])])
            leftCoord=self.minVarLen*math.ceil(insertion[1]/self.minVarLen)-self.minVarLen
            leftCoordSimple=int(math.floor(leftCoord/self.minVarLen))
            rightCoord=self.minVarLen*math.floor(insertion[1]/self.minVarLen)+self.minVarLen
            rightCoordSimple=int(math.floor(rightCoord/self.minVarLen))
            if (readChrom in self.genomeCoverage.keys() and
                leftCoordSimple<len(self.genomeCoverage[readChrom])):
                leftCov=self.genomeCoverage[readChrom][leftCoordSimple]#[0]
            else:
                leftCov=0
            if (readChrom in self.genomeCoverage.keys() and
                rightCoordSimple<len(self.genomeCoverage[readChrom])):
                rightCov=self.genomeCoverage[readChrom][rightCoordSimple]#[0]
            else:
                rightCov=0
            errPerc=int(round(insertion[4]*100/readAlignedLen,0))
            out.extend([str(insertion[2]),
                        insertion[5],
                        '1',str(max([leftCov,
                                        rightCov])),
                        str(stat.median(insertion[3])),
                        str(errPerc),
                        insertion[6]])
            outFiles[1].write('\t'.join(out)+'\n')
        self.thsize-=1

if __name__=='__main__':        

    # Section of reading arguments
    par=argparse.ArgumentParser(description='This script calls large rearrangements in ONT data')
    par.add_argument('--bam-file','-bam',
                     dest='bamFile',type=str,
                     help='BAM file with NGS reads',
                     required=True)
    par.add_argument('--chromosome','-chr',
                     dest='chrom',type=str,
                     help='chromosome to analyze. Default: all chromosomes',
                     required=False)
    par.add_argument('--start','-start',
                     dest='start',type=int,
                     help='Start coordinate for extracting reads. Default: whole chromosome or whole genome',
                     required=False,default=None)
    par.add_argument('--end','-end',
                     dest='end',type=int,
                     help='End coordinate for extracting reads. Default: whole chromosome or whole genome',
                     required=False,default=None)
    par.add_argument('--minimal-length','-len',
                     dest='minVarLen',type=int,
                     help='minimal acceptable length of variant (for InDels). Default: 50',
                     required=False,default=50)
    par.add_argument('--minimal-clipped-length','-clip',
                     dest='minClipLen',type=int,
                     help='minimal acceptable length of a clipped read part. Default: 100',
                     required=False,default=100)
    par.add_argument('--dist-to-join-trl','-dist',
                     dest='distToJoinTrl',type=int,
                     help='maximal acceptable distance between two coordinates. Default: 1000',
                     required=False,default=1000)
    par.add_argument('--threads','-th',
                     dest='threads',type=int,
                     help='number of threads to call LRs. Default: 4',
                     required=False,default=4)
    #par.add_argument('--output-file','-out',
    #                 dest='outFile',type=str,
    #                 help='VCF-file for output',
    #                 required=False)
    args=par.parse_args()
    
    ontc=ONTLRCaller(args.bamFile,args.chrom,
                     args.start,args.end,args.threads,
                     args.minVarLen,args.minClipLen,
                     args.distToJoinTrl)
    ontc.readBamFile()
    
    



from read_bam_file import BamReader
from ONTLRcaller import ONTLRCaller
from joinONTLRs import JoinLR
from alignment_INS import InsertionProcessor
from define_type_create_vcf_LRs import AnalyzeLR
import argparse
from multiprocessing import cpu_count
import glob
import os
import sys
import gc
from multiprocessing import set_start_method
import logging



def main():
    par = argparse.ArgumentParser(description='This script calls large rearrangements in ONT data')

    # if you need to continue from certain stage
    par.add_argument('--continue', '-cont',
                    dest='cont', type=str,
                    help='start the programme from a certain stage: bam - to start from ONTLRcaller, join - joinONTLRs, def - define_type_create_vcf. Default: all (for all stages)',
                    default='all')

    # ONTLRcaller
    par.add_argument('--workdir', '-dir',
                    dest='workDir', type=str,
                    help='full path to working directory',
                    required=True)
    par.add_argument('--bam-file', '-bam',
                    dest='bamFile', type=str,
                    help='full path to BAM file with NGS reads',
                    required=True)
    par.add_argument('--div-length', '-dvlen',
                    dest='len_division', type=int, default=0,
                    help='Length of regions for division chromosome to analyze',
                    required=False)
    par.add_argument('--minimal-length', '-len',
                    dest='minVarLen', type=int,
                    help='minimal acceptable length of variant (for InDels). Default: 50',
                    required=False, default=50)
    par.add_argument('--minimal-clipped-length', '-clip',
                    dest='minClipLen', type=int,
                    help='minimal acceptable length of a clipped read part. Default: 100',
                    required=False, default=100)
    par.add_argument('--dist-to-join-trl', '-dist',
                    dest='distToJoinTrl', type=int,
                    help='maximal acceptable distance between two coordinates. Default: 1000',
                    required=False, default=1000)
    par.add_argument('--threads', '-th',
                    dest='threads', type=int,
                    help='number of threads to call LRs. Default: 4',
                    required=False, default=4)
    
    # joinONTLRs
    par.add_argument('--input-files', '-in',
                    dest='inFiles', type=str,
                    help='regular expression for CSV-files with called LRs (if you have changed files names)',
                    required=False)
    par.add_argument('--maximal-distance-join', '-join',
                    dest='maxDistToJoin', type=int,
                    help='maximal acceptable distance of two neighboring fusions to join them. Default: 30',
                    required=False, default=30)
    par.add_argument('--output-lrs', '-lrs',
                    dest='outLrsFile', type=str,
                    help='CSV output file for other large rearrangements (if you want special names)',
                    required=False)
    par.add_argument('--output-ins', '-ins',
                    dest='outInsFile', type=str,
                    help='CSV output file for insertions (if you want special names)',
                    required=False)
    
    # alignment INS
    par.add_argument('--ref-genome', '-ref',
                    dest='refGen', help='path to the reference genome',
                    required=True)
    par.add_argument('--name', '-n', help='Name of INS alignment run', 
                    dest='nameINS', required=False)
    par.add_argument('--sam_file', '-sam',
                    dest='samFile', help='path to the SAM-file if mapping has already been done',
                    required=False)
    par.add_argument("--not_remove_trash_align", "-nrt_ins",
                    dest='notRemoveTrashAlign', action='store_true', required=False,
                    help="Use this parameter if you don't want to remove temporary files after INS alignment (.fasta)",
                    default=False)
    
    # define type
    par.add_argument("--junc_file", "-juncf", 
                    dest='juncFile', type=str, required=False,
                    help="full path to the file with all found genomic rearrangements after the alignment of insertions")
    par.add_argument("--ins_file", "-insf",
                    dest='insFile', type=str, required=False,
                    help="full path to the file with all found insertions after the alignment of insertions")
    par.add_argument("--out_vcf", "-out",
                    dest='outVCF', type=str, required=False,
                    help="full path to the output vcf-file")
    par.add_argument("--vcf_anno", "-vcfanno",
                    dest='VCFanno', type=str, required=True,
                    help="full path to the vcfanno binary file")
    par.add_argument("--bed_file", "-bed",
                    dest='bedFile', type=str, required=True,
                    help="full path to the bed-file with repeats and other genome elements for program annotation")
    par.add_argument("--not_remove_trash_anno", "-nrt_anno",
                    dest='notRemoveTrashAnno', action='store_true', required=False,
                    help="use this parameter if you don't want to remove temporary files (.lua, .toml, directories with vcfanno results)",
                    default=False)
    
    args=par.parse_args()

    if not os.path.isdir(args.workDir):
        os.makedirs(args.workDir, exist_ok=True)        

    logging_path = args.workDir+'/elarodon.log'
    logging.basicConfig(level=logging.DEBUG, filename=logging_path, filemode="w",
                            format="%(asctime)s %(funcName)s %(lineno)d %(message)s")
    
    logging.info("--- The eLaRodON has been started! ---")

    def check_path(path, arg_name):
        if not os.path.exists(path):
            logging.error("ERROR: " + arg_name + " does not exist: "+ path)
            raise FileNotFoundError("ERROR: " + arg_name + " does not exist: "+ path)
        return path

    if args.workDir[-1]=='/':
        args.workDir=args.workDir[:-1]

    try:
        check_path(args.bamFile, "--bam-file")
        check_path(args.refGen, "--ref-genome")
        check_path(args.VCFanno, "--vcf_anno")
        check_path(args.bedFile, "--bed_file")

        if not os.path.isdir(args.workDir):
            os.makedirs(args.workDir, exist_ok=True)
            logging.info("NOTE: Created working directory at " + args.workDir)
            print("NOTE: Created working directory at " + args.workDir)
        if not os.path.isdir(os.path.join(args.workDir, 'supplementary')):
            os.makedirs(os.path.join(args.workDir, 'supplementary'), exist_ok=True)
            
    except FileNotFoundError as e:
        print("ERROR: " + str(e))
        logging.error("ERROR: " + str(e))
        sys.exit(1)
    except Exception as e:
        print("ERROR: " + str(e))
        logging.error('ERROR: ' + str(e))
        sys.exit(1)

    bam_file_name_full = args.bamFile.split('/')[-1]
    bam_file_name = bam_file_name_full[:bam_file_name_full.rfind('.')]
    if bam_file_name[-3:] == 'bam':
        bam_file_name = bam_file_name[:-4]
    bam_results_path = args.workDir +'/supplementary/' + bam_file_name 

    logging.info('bam_file_name: ' + bam_file_name)
    logging.info('bam_results_path: ' + bam_results_path)

    if not hasattr(args, 'inFiles') or args.inFiles is None:
        args.inFiles = bam_results_path+'.junction_stat.*.*ions.csv'
    if not hasattr(args, 'outLrsFile') or args.outLrsFile is None:
        args.outLrsFile = bam_results_path+'.junction_stat.LRs_join100.csv'
    if not hasattr(args, 'outInsFile') or args.outInsFile is None:
        args.outInsFile = bam_results_path+'.junction_stat.INS_join100.csv'
    if not hasattr(args, 'nameINS') or args.nameINS is None:
        args.nameINS = bam_file_name
    if not hasattr(args, 'juncFile') or args.juncFile is None:
        args.juncFile = bam_results_path +'.junction_stat.LRs_join100_mapped_ins.csv'
    if not hasattr(args, 'insFile') or args.insFile is None:
        args.insFile = bam_results_path +'.junction_stat.INS_join100_mapped_ins.csv'
    if not hasattr(args, 'outVCF') or args.outVCF is None:
        args.outVCF = args.workDir + '/' + bam_file_name + '_all_LGRS.vcf'
    
    logging.info('args.inFiles: ' + args.inFiles)
    logging.info('args.outLrsFile: ' + args.outLrsFile)
    logging.info('args.outInsFile: ' + args.outInsFile)
    logging.info('args.nameINS: ' + args.nameINS)
    logging.info('args.juncFile: ' + args.juncFile)
    logging.info('args.insFile: ' + args.insFile)
    logging.info('args.outVCF: ' + args.outVCF)

    logging.basicConfig(level=logging.DEBUG, filename=logging_path, filemode="a",
                            format="%(asctime)s %(funcName)s %(lineno)d %(message)s")

    if args.cont == 'all' or args.cont == 'bam':
        # ONTLRcaller
        print()
        print('=== ONTLRcaller ===')
        logging.info('--- ONTLRcaller started ---')
        set_start_method('spawn')
        bamreader = BamReader(args.bamFile, args.threads, logging_path)
        chromosomes_dict = bamreader.read_bam()
        # we analyze only main chomosomes for human genome
        chom_list = set(chromosomes_dict.keys())
        chrom_intersection_v1 = chom_list & {'chr1', 'chr2', 'chr3', 'chr4', 'chr5', 'chr6', 'chr7', 'chr8', 'chr9', 'chr10', 'chr11', 'chr12', 'chr13', 'chr14', 'chr15', 'chr16', 'chr17', 'chr18', 'chr19', 'chr20', 'chr21', 'chr22', 'chrX', 'chrY', 'chrM'}
        chrom_intersection_v2 = chom_list & {'1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13', '14', '15', '16', '17', '18', '19', '20', '21', '22', 'X', 'Y', 'M'}
        if len(chrom_intersection_v1) == 25:
            chromosomes_dict = {k: v for k, v in chromosomes_dict.items() if k in chrom_intersection_v1}
        elif len(chrom_intersection_v2) == 25:
            chromosomes_dict = {k: v for k, v in chromosomes_dict.items() if k in chrom_intersection_v2}
        for chrom, chrom_len in sorted(chromosomes_dict.items()):
            if args.len_division == 0:
                ontc=ONTLRCaller(args.bamFile,
                                logging_path,
                                chrom,
                                None,
                                None,
                                args.threads,
                                args.minVarLen,
                                args.minClipLen,
                                args.distToJoinTrl,
                                args.workDir +'/supplementary/')
                ontc.readBamFile()
            else:
                n = args.len_division
                for start in range(1, chrom_len, n):
                    end = min(start + n-1, chrom_len)
                    ontc=ONTLRCaller(args.bamFile,
                                    logging_path,
                                    chrom,
                                    start,
                                    end,
                                    args.threads,
                                    args.minVarLen,
                                    args.minClipLen,
                                    args.distToJoinTrl,
                                    args.workDir +'/supplementary/')
                    ontc.readBamFile()
        logging.info('--- ONTLRcaller finished ---')
        del(ontc)
        gc.collect()
    
    if args.cont == 'all' or args.cont == 'bam' or args.cont == 'join':
        #joinONTLRs
        print()
        print('=== joinONTLRs ===')
        logging.info('--- joinONTLRs started ---')
        ds = glob.glob(args.inFiles)
        if len(ds)==0:
            print('ERROR (1)! No files were chosen:')
            logging.error('ERROR (1)! No files were chosen:')
            print(args.inFiles)
            logging.error('args.inFiles: ', args.inFiles)
            exit(1)
        print('The number of files chosen:', len(ds))
        logging.info('The number of files chosen: '+str(len(ds)))
        print('Reading input files with', args.threads, 'threads...')
        logging.info('Reading input files with '+str(args.threads)+' threads...')
        jl = JoinLR(ds, args)
        print('The total number of fusions:', sum(len(item) for item in jl.allFusions.values()))
        logging.info('The total number of fusions: '+str(sum(len(item) for item in jl.allFusions.values())))
        print('The total number of insertions:', sum(len(item) for item in jl.allInsertions.values()))
        logging.info('The total number of insertions: '+str(sum(len(item) for item in jl.allInsertions.values())))
        jl.joinAllSimilarLRs()
        logging.info('jl.joinAllSimilarLRs()')
        jl.writeFusionToOutput(args.outLrsFile)
        logging.info('jl.writeFusionToOutput(args.outLrsFile)')
        jl.writeInsertionsToOutput(args.outInsFile)
        logging.info('jl.writeInsertionsToOutput(args.outInsFile)')
        del(jl)
        gc.collect()
        logging.info('--- joinONTLRs finished ---')
        print()

        # alignment INS
        print()
        print('=== alignment_INS ===')
        logging.info('--- alignment_INS started ---')
        alignINS_processor = InsertionProcessor(logging_path,
                                                args.outInsFile, 
                                                args.outLrsFile, 
                                                args.refGen, 
                                                args.nameINS, 
                                                args.samFile,
                                                args.threads, 
                                                args.notRemoveTrashAlign)
        alignINS_processor.read_insertions()
        fasta_file = args.workDir +'/supplementary/'+ args.nameINS + '_insertions.fasta'
        alignINS_processor.write_fasta(fasta_file)
        alignINS_processor.map_with_minimap2(fasta_file)
        alignINS_processor.analyze_mapping_results()
        alignINS_processor.merge_tandem_duplications()
        alignINS_processor.copy_file(args.outLrsFile, args.outLrsFile[:-4]+'_mapped_ins.csv')
        alignINS_processor.extract_sequences_before_after()
        alignINS_processor.update_large_rearrangements_file()
        updated_csv = args.outInsFile[:-4]+'_mapped_ins.csv'
        alignINS_processor.update_original_csv(updated_csv)
        del(alignINS_processor)
        gc.collect()
        print('Analysis of INS was done!')
        logging.info('--- alignment_INS finished ---')

    if args.cont == 'all' or args.cont == 'bam' or args.cont == 'join' or args.cont == 'def':
        # define types of LGRs
        print()
        print('=== define_type_create_vcf_LRs ===')
        logging.info('--- define_type_create_vcf_LRs started ---')
        analyze_LR = AnalyzeLR(args=args)
        analyze_LR.main_func()
        logging.info('--- define_type_create_vcf_LRs finished ---')
    
    logging.info("--- The eLaRodON has been finished! ---")
    

if __name__ == '__main__':
    main()